module Blocks where

import qualified OP
import qualified IR

data BasicBlock
     = BB { entryLabels :: [IR.Label]
          , exitLabels  :: [IR.Label]
          , statements  ::  [IR.Stm]
          }
     deriving Show
     -- entry labels, exit labels, statments

basicBlocks :: [IR.Stm] -> [BasicBlock]
basicBlocks [] = []
basicBlocks ss = let (block, rest) = splitBasic [] [] ss in block : basicBlocks rest

splitBasic entries acc (s@(IR.LABEL i) : rest) = splitBasic (i : entries) (s : acc) rest
splitBasic entries acc ss = splitBasic1 entries acc ss

splitBasic1 ins acc (s@(IR.JUMP lt) : rest) = (BB ins [lt] $ reverse (s : acc), rest)
splitBasic1 ins acc (s@(IR.CJUMP _ _ _ lt lf) : rest) = (BB ins [lt, lf] $ reverse (s : acc), rest)
splitBasic1 ins acc rest@((IR.LABEL le) : _) = (BB ins [le] $ reverse acc, rest)
splitBasic1 ins acc (s : rest) = splitBasic1 ins (s : acc) rest
splitBasic1 ins acc ss = (BB ins [] $ reverse acc, ss)

type Trace = [IR.Stm]

traces :: [BasicBlock] -> [Trace]
traces [] = []
traces (bb : bbs) =
    let (trace, rest) = splitTrace [bb] bbs
    in  flattenTrace trace : traces rest

splitTrace acc [] = (reverse acc, [])
splitTrace acc@(BB ins outs _ : _) bbs = 
    let (mb, rest) = extractBB outs bbs in
    case mb of
      Just bb -> splitTrace (bb : acc) rest
      Nothing -> (reverse acc, rest)

extractBB outs [] = (Nothing, [])
extractBB outs (bb@(BB ins _ _) : bbs) = 
    if any (`elem` outs) ins then (Just bb, bbs) else 
    let (mb, rest) = extractBB outs bbs in (mb, bb : rest)

flattenTrace :: [BasicBlock] -> Trace
flattenTrace bbs = concatMap statements bbs

-- 

traceOpt :: Trace -> Trace
traceOpt (IR.JUMP t : rest@(IR.LABEL t': _)) 
    | t == t' = traceOpt rest
traceOpt (IR.CJUMP op e1 e2 lt lf : rest@(IR.LABEL t : _)) 
    | lt == t = IR.CJUMP (OP.invert op) e1 e2 lf lt : traceOpt rest
traceOpt (s : ss)
    = s : traceOpt ss
traceOpt [] 
    = []
