{-# LANGUAGE TypeFamilies, FlexibleInstances #-}
module MaxMunge where

import OP
import qualified IR
import Temp
import qualified Instr

class GenAsm ir where
  type F ir 
  gen :: TMonad m => ir -> m (F ir)

instance GenAsm IR.Exp where
  type F IR.Exp = (IR.Temp, [Instr.T])
  gen (IR.TEMP r) =
      return (r, [])
  gen (IR.CONST c) = do
      rt <- newTemp
      return (rt, [Instr.OPER ("Addi 'd0,$0," ++ show c) [rt] []])
  gen (IR.BINOP Add (IR.CONST c) t) = do
      rt <- newTemp
      (r, is) <- gen t
      return (rt, is ++ [Instr.OPER ("Addi 'd0,'s0," ++ show c) [rt] [r]])
  gen (IR.BINOP Add t (IR.CONST c)) = do
      rt <- newTemp
      (r, is) <- gen t
      return (rt, is ++ [Instr.OPER ("Addi 'd0,'s0," ++ show c) [rt] [r]])
  gen (IR.BINOP Sub t (IR.CONST c)) = do
      rt <- newTemp
      (r, is) <- gen t
      return (rt, is ++ [Instr.OPER ("Subi 'd0,'s0," ++ show c) [rt] [r]])
  gen (IR.BINOP op t1 t2) = do
      rt <- newTemp
      (r1, is1) <- gen t1
      (r2, is2) <- gen t2
      return (rt, is1 ++ is2 ++ [Instr.OPER (show op ++ " 'd0,'s0,'s1") [rt] [r1, r2]])
  gen (IR.MEM (IR.BINOP Add (IR.CONST c) t)) = do
      rt <- newTemp
      (r, is) <- gen t
      return (rt, is ++ [Instr.OPER ("Load 'd0," ++ show c ++ "('s0)") [rt] [r]])
  gen (IR.MEM (IR.BINOP Add t (IR.CONST c))) = do
      rt <- newTemp
      (r, is) <- gen t
      return (rt, is ++ [Instr.OPER ("Load 'd0," ++ show c ++ "('s0)") [rt] [r]])
  gen (IR.MEM (IR.CONST c)) = do
      rt <- newTemp
      return (rt, [Instr.OPER ("Load 'd0," ++ show c ++ "($0)") [rt] []])
  gen (IR.MEM t) = do
      rt <- newTemp
      (r, is) <- gen t
      return (rt, is ++ [Instr.OPER "Load 'd0,0('s0)" [rt] [r]])
  gen (IR.ESEQ stms exp) = do
      is1 <- gen stms
      (rt, is2) <- gen exp
      return (rt, is1 ++ is2)

instance GenAsm IR.Stm where
  type F IR.Stm = [Instr.T]
  gen (IR.EXP e) = do
      (r, is) <- gen e
      return is
  gen (IR.MOVE (IR.MEM e1) (IR.MEM e2)) = do
      (r1, is1) <- gen e1
      (r2, is2) <- gen e2
      return (is1 ++ is2 ++ [Instr.OPER "Movem 'd0,'s0" [r1] [r2]])
  gen (IR.MOVE (IR.MEM (IR.BINOP Add e1 (IR.CONST c))) e2) = do
      (r1, is1) <- gen e1
      (r2, is2) <- gen e2
      return (is1 ++ is2 ++ [Instr.OPER ("Store " ++ show c ++ "('d0),'s0") [r1] [r2]])
  gen (IR.MOVE (IR.MEM (IR.BINOP Add (IR.CONST c) e1)) e2) = do
      (r1, is1) <- gen e1
      (r2, is2) <- gen e2
      return (is1 ++ is2 ++ [Instr.OPER ("Store " ++ show c ++ "('d0),'s0") [r1] [r2]])
  gen (IR.MOVE (IR.MEM (IR.CONST c)) e2) = do
      (r2, is2) <- gen e2
      return (is2 ++ [Instr.OPER ("Store " ++ show c ++ "($0),'s0") [] [r2]])
  gen (IR.MOVE (IR.MEM e1) e2) = do
      (r1, is1) <- gen e1
      (r2, is2) <- gen e2
      return (is1 ++ is2 ++ [Instr.OPER "Store 0('d0),'s0" [r1] [r2]])
  gen (IR.LABEL l) =
      return ([Instr.LABEL l])
  gen (IR.JUMP l) =
      return ([Instr.OPER ("Jump L" ++ show l) [] []])
  gen (IR.CJUMP rop e1 e2 lt lf) = do
      (r1, is1) <- gen e1
      (r2, is2) <- gen e2
      return (is1 ++ is2 ++ [Instr.OPER "Compare 's0,'s1" [] [r1,r2], Instr.OPER ("J" ++ show rop ++ " L" ++ show lt) [] []])

instance GenAsm [IR.Stm] where
  type F [IR.Stm] = [Instr.T]
  gen [] = 
      return []
  gen [ir] =
      gen ir
  gen (ir:irs) = do
      is1 <- gen ir
      is2 <- gen irs
      return (is1 ++ is2)
